from scipy import *
from scipy import stats
import shelve,pdb, ctypes,random
from SpCAError import *

class SpCAMI:
    def __init__(self,gaptol=0.1):
        self.AADic={'A':0,'C':1,'D':2,'E':3,'F':4,'G':5,'H':6,'I':7,'K':8,'L':9,'M':10,'N':11,'P':12,'Q':13,'R':14,'S':15,'T':16,'V':17,'W':18,'Y':19}
        self.AlignDic = {}; self.AlignSeqIds=[]; self.seqnum = 0; self.seqlen = 0
        self.gaptol=gaptol
        self.minfloat32 = finfo(float32).tiny.tolist()
        self.minfloat64 = finfo(float64).tiny.tolist()
        self.K1 = zeros((20,20),dtype='float')
        self.K2 = zeros((400,400),dtype='float')
    def readAlign(self,alifn):
        alifp = open(alifn)
        self.seqnum = 0
        self.seqlen = 0
        for oneline in alifp:
            oneline=oneline.strip()
            id, seq = oneline.split()
            self.AlignDic[id]=seq
            self.AlignSeqIds.append(id)
            self.seqnum = self.seqnum+1
            self.seqlen = len(seq)
        #|
        alifp.close()
        return self.AlignDic
    def setAlignf(self,alifn):
        seqdic = self.readAlign(alifn)
        return seqdic
    def read1col(self,colnum):
        container=[]
        for id in self.AlignSeqIds:
            aa=self.AlignDic[id][colnum]
            container.append(aa)
        #| for id
        return container    
    def criteria(self, col_list):
        numOfgaps = col_list.count("-")
        if float(numOfgaps)/self.seqnum > self.gaptol:
            raise CriteriaError, "gap tolerance error"
        #|
        return True    
    def rtSubMat1col(self,colnum):
        onecol = self.read1col(colnum)
        if not self.criteria(onecol):
            return False
        #|
        length = len(onecol)
        container = []
        for m in xrange(1,length):
            container.append((onecol[0],onecol[m]))
        #|
        return container
    def rtIdx(self,pair):
        aapair=''.join(pair)
        try:
            return self.AAPairDic[aapair]
        except:
            raise NoKeyErr, aapair
    def countKMat1(self, i,j,randomf=False):
        self.K1 *=0
        coli = self.read1col(i)
        colj = self.read1col(j)
        try:
            self.criteria(coli); self.criteria(colj)
        except CriteriaError,e:
            raise CriteriaError,e
        #|
        if randomf:
            random.shuffle(colj); random.shuffle(coli)
        for aai, aaj in zip(coli,colj):
            try:
                idxi = self.AADic[aai]
                idxj = self.AADic[aaj]
                self.K1[idxi,idxj]+=1
            except KeyError:
                continue
        #|
        return self.K1
    def calMI(self,K,probKfl=False):
        if not probKfl:
            tot_sum = K.sum()
            Pij = array(K.tolist(),dtype='f')/tot_sum
        else:
            Pij = K
        #|
        pi = array(map(sum,Pij),dtype='f')
        pj = array(map(sum,Pij.T),dtype='f')
        pipj=array(matrix(pi).T*matrix(pj))
        logM=log2(Pij/pipj)
        logM=logM/log2(20)
        idces=Pij.nonzero()
        resM=Pij[idces]*logM[idces]
        return resM.sum()
    def calH(self,K,direction):
        tot_sum = K.sum()
        Pij = array(K.tolist(),dtype='f')/tot_sum
        if direction == "Hx":
            p = array(map(sum,Pij),dtype='f')
        elif direction == "Hy":
            p = array(map(sum,Pij.T),dtype='f')
        elif direction == "Hxy":
            p = Pij.flatten()
        #|
        logM = log2(p); logM = logM/log2(20)
        idces = p.nonzero()
        resM = p[idces]*logM[idces]
        return -resM.sum()
    def popKeyRes(self, key, idx):
        return self.AlignDic[key][idx]

if __name__=="__main__":
    import sys
    from optparse import OptionParser
    def main():
        getopt = OptionParser('usage: %prog [options] fodor.aln')
        getopt.add_option('-q','--query',action='store',dest='queryId',help="Query Id")
        getopt.add_option('-G','--gaptol',action='store',dest='gaptolerance',metavar='ratio[0~1.0] default: 0.0', help="criteria: no calculation > gapratio")
        getopt.set_default("gaptolerance",0.0)
        getopt.add_option('-v','--version',type='int',action='store',dest='version',metavar="(Required) [1,2]",help='1: MI, 2: Tillier Entropy Factor HiHj(1-HiHj)')
        getopt.add_option('--initidx',action='store',type='int',dest='initidx',help="initial index instead of 0")
        getopt.set_default("initidx",0)
        options,argv = getopt.parse_args()
        if options.version==None or options.queryId==None or len(argv)!=1:
            getopt.print_help()
            sys.exit(1)
        #| if
        mi = SpCAMI(gaptol=float(options.gaptolerance))
        mi.setAlignf(argv[0])
        N=mi.seqlen
        queryId=options.queryId
        sys.stdout.write("i\tj\tscore\tAAi\tAAj\n")
        for i in xrange(N-1):
            for j in xrange(i+1,N):
                try:
                    if options.version==1:
                        K1 = mi.countKMat1(i,j,randomf=False)
                        score = mi.calMI(K1)
                    elif options.version==2:
                        K1 = mi.countKMat1(i,j,randomf=False)
                        Hx = mi.calH(K1,'Hx')
                        Hy = mi.calH(K1,'Hy')
                        score = Hx*Hy*(1-Hx*Hy)
                    #| if
                    sys.stdout.write("%d\t%d\t%f\t%s\t%s\n" %(i+options.initidx,j+options.initidx,score,mi.popKeyRes(queryId,i),mi.popKeyRes(queryId,j)))
                except Exception, e:
                    sys.stderr.write("%d\t%d\t%s\n" %(i+options.initidx,j+options.initidx,e))
                #| tr
            #| for j
        #| for i
    #| main
    main()
