from scipy import *
from scipy import stats
import copy

class IterCoVar:
    def __init__(self):
        self.thresval = False
        self.resDic = {}
    def mkS(self,ks):
        s = set([])
        for i,j in ks:
            s.add(i); s.add(j)
        return s
    def mkIdx(self,ks):
        s = self.mkS(ks)
        l = len(s)
        return range(l)
    def mkIdxDic(self, ks):
        s = self.mkS(ks)
        l = self.mkIdx(ks)
        s = list(s)
        s.sort()
        Dic = {}
        for i,j in zip(s,l):
            Dic[i]=j
        #|
        return Dic
    def sigmoid(self,X):
        m = X.mean()
        return 1.0/(1.0+exp(-10.0*(X-m)))
    def readCMAf(self, fn,sigmoidf=False):
        fp = open(fn)
        KDic={}
        for l in fp:
            i,j,score,aai,aaj = l.split()
            try:
                self.resDic[int(i)]=aai
                self.resDic[int(j)]=aaj
                KDic[(int(i),int(j))]=float(score)
            except:
                continue
        ks=KDic.keys()
        self.Kidx = self.mkIdx(ks)
        self.KidxDic = self.mkIdxDic(ks)
        K = zeros((len(self.Kidx),len(self.Kidx)),dtype='f')
        for i,j in KDic:
            i_Kidx = self.KidxDic[i]; j_Kidx = self.KidxDic[j]
            K[i_Kidx,j_Kidx] = K[j_Kidx,i_Kidx] = KDic[(i,j)]
        #|
        self.CMAMaxVal = K.max()
        K[self.Kidx,self.Kidx]=1.0
        if sigmoidf:
            K = self.sigmoid(K)
        else:
            pass
        self.CMAK = K.tolist()
        return K
    def rtGlobVecs(self, idxi, idxj,thr_cut_ratio=1.0):
        ori_i = copy.copy(self.CMAK[self.KidxDic[idxi]])
        ori_j = copy.copy(self.CMAK[self.KidxDic[idxj]])
        ref_i = []; ref_j = []
        if thr_cut_ratio==1.0:
            minimum = min([idxi,idxj]); maximum = max([idxi,idxj])
            ori_i.pop(self.KidxDic[minimum]); ori_j.pop(self.KidxDic[minimum])
            ori_i.pop(self.KidxDic[maximum]-1); ori_j.pop(self.KidxDic[maximum]-1)
            return ori_i,ori_j
        else:
            pass
        thresval = self.rtThresholdVal(thr_cut_ratio)
        for i in xrange(len(ori_i)):
            if i != self.KidxDic[idxi] and i != self.KidxDic[idxj]:
                if ori_i[i]>=thresval:
                    ref_i.append(ori_i[i])
                else:
                    ref_i.append(0.0)
                if ori_j[i]>=thresval:
                    ref_j.append(ori_j[i])
                else:
                    ref_j.append(0.0)
                #| if
            #| if
        #| for
        return ref_i,ref_j
    def similarity(self,idxi,idxj,type="CPS",cutoffratio=1.0):
        if type=="CPS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_method2(veci,vecj)
        else:
            raise Exception, "No method:"+type
        #|
        return r
    def rtCMAKval(self,i,j):
        idxi = self.KidxDic[i]
        idxj = self.KidxDic[j]
        return self.CMAK[idxi][idxj]
    def rtCMAMaxVal(self):
        return self.CMAMaxVal
    def sim_method2(self, veci, vecj):
        ai = array(veci)
        aj = array(vecj)
        prod = ai*aj
        nnonzero = len(prod[prod.nonzero()])
        try:
            if nnonzero!=0:
                res = prod.sum()/nnonzero
            else:
                res = 0
        except Exception, e:
            sys.stderr.write("%d,%d\n" %(len(ai),len(aj)))
            raise Exception,e
        return res
    def rtTotalMean(self,sqrtq=False):
        if sqrtq:
            A = sqrt(array(self.CMAK))
        else:
            A=array(self.CMAK)
        #|
        N = len(self.CMAK)
        res = A.sum()-1.0*N
        res = res/(N*(N-1))
        return res
    def setPPSMat(self,method,cutoffratio=1.0):
        idxlist = self.KidxDic.keys()
        idxlist.sort()
        container = zeros((len(idxlist),len(idxlist)),dtype='f')
        for ii in xrange(len(idxlist)-1):
            idxi = idxlist[ii]
            for jj in xrange(ii+1,len(idxlist)):
                idxj = idxlist[jj]
                pps = self.similarity(idxi,idxj,type=method,cutoffratio=cutoffratio)
                if ii==jj:
                    continue
                else:
                    container[self.KidxDic[idxi],self.KidxDic[idxj]]=pps
                    container[self.KidxDic[idxj],self.KidxDic[idxi]]=pps
            #|
        #|
        return container
    def sumPPSMat(self,ppsmat):
        return ppsmat.sum()
    def sumPPSMatij(self,ppsmat,idxi,idxj):
        i = self.KidxDic[idxi]; j = self.KidxDic[idxj]
        return ppsmat[i,:].sum()+ppsmat[j,:].sum()+ppsmat[:,i].sum()+ppsmat[:,j].sum()-2*ppsmat[i,j]
    def rtNormFact(self,ppsmat,sqrtq=True):
        n = len(ppsmat)
        m = n*(n-1)
        if sqrtq:
            return sqrt(self.sumPPSMat(ppsmat)/m)
        else:
            return self.sumPPSMat(ppsmat)/m
    def popRes(self,idx):
        return self.resDic[idx]

if __name__=="__main__":
    import sys
    from optparse import OptionParser
    getopt = OptionParser('usage: %prog MI.out entfact.out')
    getopt.add_option('-v','--version',action='store',type='int',dest='version',metavar="1/2/3",help='1: MIc, 2: eMIc, 3: aMIc')
    options,argv=getopt.parse_args()
    if len(argv)!=2:
        getopt.print_help()
        sys.exit(1)
    #| if
    mic = IterCoVar()
    mic.readCMAf(argv[0])
    efact = IterCoVar()
    efact.readCMAf(argv[1])
    idxlist = mic.KidxDic.keys()
    idxlist.sort()
    NoiseMat1 = mic.setPPSMat("CPS",cutoffratio=1.0)
    norm_fact1 = mic.rtNormFact(NoiseMat1)
    MIcMat = zeros(NoiseMat1.shape,dtype='f')
    efactMat = zeros(NoiseMat1.shape,dtype='f')
    print "i\tj\tscore\tAAi\tAAj"
    for i in xrange(len(idxlist)-1):
        idxi = idxlist[i]
        for j in xrange(i+1,len(idxlist)):
            idxj = idxlist[j]
            noise1 = NoiseMat1[mic.KidxDic[idxi],mic.KidxDic[idxj]]
            MIcMat[i,j] = mic.rtCMAKval(idxi,idxj)-noise1/norm_fact1 #MIc
            if options.version==1:
                sys.stdout.write("%d\t%d\t%f\t%s\t%s\n" %(idxi,idxj,MIcMat[i,j],mic.popRes(idxi),mic.popRes(idxj)))
            if options.version!=1:
                efactMat[i,j] = efact.rtCMAKval(idxi,idxj)
    if options.version==1:
        sys.exit(0)
    else:
        EMIcMat = efactMat*MIcMat
        EMIcMat = EMIcMat/EMIcMat.max()
        MIcMat = MIcMat/MIcMat.max()
    for i in xrange(len(idxlist)-1):
        idxi = idxlist[i]
        for j in xrange(i+1,len(idxlist)):
            idxj = idxlist[j]
            score2 = EMIcMat[i,j]
            if options.version==3:
                score1 = MIcMat[i,j]
                sys.stdout.write("%d\t%d\t%f\t%s\t%s\n" %(idxi,idxj,0.5*(score1+score2),mic.popRes(idxi),mic.popRes(idxj)))
            elif options.version==2:
                sys.stdout.write("%d\t%d\t%f\n" %(idxi,idxj,score2))
