from scipy import *
from scipy import stats
import copy
import unittest
class IterCoVar:
    def __init__(self):
        self.thresval = False
    def mkS(self,ks):
        s = set([])
        for i,j in ks:
            s.add(i); s.add(j)
        return s
    def mkIdx(self,ks):
        s = self.mkS(ks)
        l = len(s)
        return range(l)
    def mkIdxDic(self, ks):
        s = self.mkS(ks)
        l = self.mkIdx(ks)
        s = list(s)
        s.sort()
        Dic = {}
        for i,j in zip(s,l):
            Dic[i]=j
        #|
        return Dic
    def sigmoid(self,X):
        m = X.mean()
        return 1.0/(1.0+exp(-10.0*(X-m)))
    def readCMAf(self, fn,sigmoidf=False):
        fp = open(fn)
        KDic={}
        for l in fp:
            i,j,score = l.split()
            try:
                KDic[(int(i),int(j))]=float(score)
            except:
                continue
        ks=KDic.keys()
        self.Kidx = self.mkIdx(ks)
        self.KidxDic = self.mkIdxDic(ks)
        K = zeros((len(self.Kidx),len(self.Kidx)),dtype='f')
        for i,j in KDic:
            i_Kidx = self.KidxDic[i]; j_Kidx = self.KidxDic[j]
            K[i_Kidx,j_Kidx] = K[j_Kidx,i_Kidx] = KDic[(i,j)]
        #|
        self.CMAMaxVal = K.max()
        K[self.Kidx,self.Kidx]=1.0
        if sigmoidf:
            K = self.sigmoid(K)
        else:
            pass
        self.CMAK = K.tolist()
        return K
    def rtThresholdVal(self, percentage):
        if not self.thresval:
            tmp = array(self.CMAK).flatten().tolist()
            tmp.sort(lambda y,x: cmp(x,y)) #descending order
            if percentage==1.0: #minimum
                self.thresval = tmp[-1]
            elif percentage==0.0: #maximum
                self.thresval =  tmp[0]
            else:
                N = int(len(tmp)*percentage)
                self.thresval = tmp[N]
        #| if not ...
        return self.thresval
    def rtGlobVecs(self, idxi, idxj,thr_cut_ratio=1.0,reversq=False):
        if reversq:
            ref_i,ref_j = self.reverse_rtGlobVecs(idxi,idxj,thr_cut_ratio)
        else:
            ori_i = copy.copy(self.CMAK[self.KidxDic[idxi]])
            ori_j = copy.copy(self.CMAK[self.KidxDic[idxj]])
            ref_i = []; ref_j = []
            if thr_cut_ratio==1.0:
                minimum = min([idxi,idxj]); maximum = max([idxi,idxj])
                ori_i.pop(self.KidxDic[minimum]); ori_j.pop(self.KidxDic[minimum])
                ori_i.pop(self.KidxDic[maximum]-1); ori_j.pop(self.KidxDic[maximum]-1)
                return ori_i,ori_j
            else:
                pass
            thresval = self.rtThresholdVal(thr_cut_ratio)
            for i in xrange(len(ori_i)):
                if i != self.KidxDic[idxi] and i != self.KidxDic[idxj]:
                    if ori_i[i]>=thresval:
                        ref_i.append(ori_i[i])
                    else:
                        ref_i.append(0.0)
                    if ori_j[i]>=thresval:
                        ref_j.append(ori_j[i])
                    else:
                        ref_j.append(0.0)
                    #| if
                #| if
            #| for
        #| if reversq
        return ref_i,ref_j
    def reverse_rtGlobVecs(self, idxi, idxj,thr_cut_ratio=1.0):
        ori_i = copy.copy(self.CMAK[self.KidxDic[idxi]])
        ori_j = copy.copy(self.CMAK[self.KidxDic[idxj]])
        ref_i = []; ref_j = []
        if thr_cut_ratio==1.0:
            l = len(ori_i)-2
            ref_i = [0]*l; ref_j = [0]*l
            return ref_i,ref_j
        else:
            pass
        thresval = self.rtThresholdVal(thr_cut_ratio)
        for i in xrange(len(ori_i)):
            if i != self.KidxDic[idxi] and i != self.KidxDic[idxj]:
                if ori_i[i]<thresval:
                    ref_i.append(ori_i[i])
                else:
                    ref_i.append(0.0)
                if ori_j[i]<thresval:
                    ref_j.append(ori_j[i])
                else:
                    ref_j.append(0.0)
        return ref_i,ref_j
    def similarity(self,idxi,idxj,type="Pearson",cutoffratio=1.0,reversq=False):
        if type=="Pearson":
            veci,vecj = self.rtGlobVecs(idxi,idxj)
            r,p = stats.pearsonr(veci,vecj)
        elif type=="Overlap":
            veci,vecj = self.rtGlobVecs(idxi,idxj)
            r = self.sim_overlap(veci,vecj,cutoffratio)
        elif type=="SRP":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_method1(veci,vecj)
        elif type=="SIMPRO":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_simpro(veci,vecj)
        elif type=="PPS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio,reversq=reversq)
            r = self.sim_method2(veci,vecj)
        elif type=="Dot":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio,reversq=reversq)
            r = dot(veci,vecj)            
        elif type=="Method2.1":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_method2_1(veci,vecj)
        elif type=="APCprime":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=1.0)
            r = self.sim_apcprime(veci,vecj)
        elif type=="APC":
            r = self.sim_apc(idxi,idxj)
        elif type=="Method30.2":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=1.0)
            r = self.sim_method30_2(veci,vecj)
        elif type=="Pearsonsqrt":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r,p = stats.pearsonr(sqrt(array(veci)),sqrt(array(vecj)))
        elif type=="TPS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_tps(veci,vecj)
        elif type=="UPS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_ups(veci,vecj)            
        elif type=="TPS2.1":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_tps2_1(veci,vecj)
        elif type=="UPS2.1":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_ups2_1(veci,vecj)
        elif type=="PAMS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_pams(veci,vecj)
        elif type=="PHMS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_phms(veci,vecj)
        elif type=="TGMS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_tgms(veci,vecj)
        elif type=="TAMS":
            veci,vecj = self.rtGlobVecs(idxi,idxj,thr_cut_ratio=cutoffratio)
            r = self.sim_tams(veci,vecj)            
        else:
            raise Exception, "No method:"+type
        #|
        return r
    def sltHighRank(self, veci,R):
        a = array(veci)
        rank = a[::-1].argsort().tolist() #descending order
        container=[]
        N = int(R*len(veci))
        for i in xrange(N):
            container.append(rank.index(i))
        return set(container)
    def sim_overlap(self, veci, vecj,cutoffratio=0.2):
        hranki = self.sltHighRank(veci,cutoffratio)
        hrankj = self.sltHighRank(vecj,cutoffratio)
        n = hranki & hrankj
        N = len(hranki)
        return float(len(n))/N
    def sim_method1(self, veci, vecj):
        m = len(veci)
        ai = array(veci)
        aj = array(vecj)
        prod = ai*aj; nnonzero = len(prod[prod.nonzero()])
        try:
            res = sqrt(prod).sum()
        except Exception, e:
            sys.stderr.write("%d,%d\n" %(len(ai),len(aj)))
            raise Exception,e
        return res/nnonzero
    def sim_simpro(self, veci, vecj):
        ai = array(veci); aj = array(vecj)
        nonzeroi = ai.nonzero()
        nonzeroj = aj.nonzero()
        mi = len(ai[nonzeroi]); mj = len(ai[nonzeroj])
        if mi==0:
            indirect_ij=0.0
        else:
            indirect_ij = ai[nonzeroj].sum()/mi
        if mj == 0:
            indirect_ji=0.0
        else:
            indirect_ji = aj[nonzeroi].sum()/mj
        return 0.5*(indirect_ij+indirect_ji)
    def rtCMAKval(self,i,j):
        idxi = self.KidxDic[i]
        idxj = self.KidxDic[j]
        return self.CMAK[idxi][idxj]
    def rtCMAMaxVal(self):
        return self.CMAMaxVal
    def sim_method2(self, veci, vecj):
        ai = array(veci)
        aj = array(vecj)
        prod = ai*aj
        nnonzero = len(prod[prod.nonzero()])
        try:
            if nnonzero!=0:
                res = prod.sum()/nnonzero
            else:
                res = 0
        except Exception, e:
            sys.stderr.write("%d,%d\n" %(len(ai),len(aj)))
            raise Exception,e
        return res
    def sim_method2_1(self, veci, vecj):
        ai = array(veci)
        aj = array(vecj)
        prod = ai*aj
        m = len(prod[prod.nonzero()])
        try:
            res = prod.sum()/(m*m)
        except Exception, e:
            sys.stderr.write("%d,%d\n" %(len(ai),len(aj)))
            raise Exception,e
        return res
    def sim_apcprime(self,veci,vecj):
        return mean(veci)*mean(vecj)
    def sim_apc(self,i,j):
        idxi = self.KidxDic[i]
        idxj = self.KidxDic[j]
        veci = copy.copy(self.CMAK[idxi])
        vecj = copy.copy(self.CMAK[idxj])
        veci.pop(idxi); vecj.pop(idxj)
        #totalsum = array(self.CMAK).sum()-len(self.CMAK)
        return mean(veci)*mean(vecj)
    def rtTotalMean(self,sqrtq=False):
        if sqrtq:
            A = sqrt(array(self.CMAK))
        else:
            A=array(self.CMAK)
        #|
        N = len(self.CMAK)
        res = A.sum()-1.0*N
        res = res/(N*(N-1))
        return res
    def sim_method30_2(self,veci,vecj):
        ai = array(veci); aj = array(vecj)
        v1 = ai.sum()*aj.sum()
        v2 = (ai*aj).sum()
        N = len(veci)
        res = (v1-v2)/(N*N)
        return res
    def sim_tps(self, veci,vecj):
        si = array(veci).sum()
        sj = array(vecj).sum()
        return si*sj/len(veci)
    def sim_ups(self, veci, vecj):
        tps = self.sim_tps(veci,vecj)
        pps = self.sim_method2(veci,vecj)
        return tps-pps
    def sim_tps2_1(self, veci,vecj):
        si = array(veci).sum()
        sj = array(vecj).sum()
        m = len(veci)
        return si*sj/(m*m)
    def sim_ups2_1(self, veci, vecj):
        tps = self.sim_tps2_1(veci,vecj)
        pps = self.sim_method2_1(veci,vecj)
        return tps-pps
    def sim_pams(self, veci, vecj):
        ai = array(veci); aj = array(vecj)
        si = ai+aj
        si = pow(si/2,2)
        return si.mean()
    def sim_phms(self, veci, vecj):
        ai = array(veci); aj = array(vecj)
        numer = 2*ai*aj
        denom = ai+aj
        hms = []
        for n,d in zip(numer,denom):
            if d!=0:
                hms.append(n/d)
            else:
                hms.append(0)
            #|
        #|
        hms = pow(array(hms),2)
        return hms.mean()
    def sim_tgms(self, veci, vecj):
        li = log(array(veci))
        lj = log(array(vecj))
        m = len(veci)
        return exp((li+lj).sum()/m)
    def sim_tams(self, veci, vecj):
        ai = array(veci)
        aj = array(vecj)
        m = 2*(len(veci))
        return pow((ai.sum()+aj.sum())/m,2)
    def setPPSMat(self,method,cutoffratio=1.0):
        idxlist = self.KidxDic.keys()
        idxlist.sort()
        container = zeros((len(idxlist),len(idxlist)),dtype='f')
        for ii in xrange(len(idxlist)-1):
            idxi = idxlist[ii]
            for jj in xrange(ii+1,len(idxlist)):
                idxj = idxlist[jj]
                pps = self.similarity(idxi,idxj,type=method,cutoffratio=cutoffratio)
                if ii==jj:
                    continue
                else:
                    container[self.KidxDic[idxi],self.KidxDic[idxj]]=pps
                    container[self.KidxDic[idxj],self.KidxDic[idxi]]=pps
            #|
        #|
        return container
    def sumPPSMat(self,ppsmat):
        return ppsmat.sum()
    def sumPPSMatij(self,ppsmat,idxi,idxj):
        i = self.KidxDic[idxi]; j = self.KidxDic[idxj]
        return ppsmat[i,:].sum()+ppsmat[j,:].sum()+ppsmat[:,i].sum()+ppsmat[:,j].sum()-2*ppsmat[i,j]
    def rtNormFact(self,ppsmat,sqrtq=True):
        n = len(ppsmat)
        m = n*(n-1)
        if sqrtq:
            return sqrt(self.sumPPSMat(ppsmat)/m)
        else:
            return self.sumPPSMat(ppsmat)/m
    def rtNormFactp(self,ppsmat,idxi,idxj):
        ppssum = self.sumPPSMatij(ppsmat,idxi,idxj)
        n = len(ppsmat); m = 4*(n-1)-2
        return sqrt(ppssum/m)
    def rtNormFactpp(self,ppsmat,idxi,idxj):
        ttsum = self.sumPPSMat(ppsmat)
        ttsum_dij = ttsum - self.sumPPSMatij(ppsmat,idxi,idxj)
        n = len(ppsmat)
        mval = ttsum_dij/((n-2)*(n-3))
        return sqrt(mval)
    def rtNormFactGM(self,noisemat):
        lm = log(noisemat)
        idxil = []; idxjl = []
        for i in xrange(len(noisemat)):
            for j in xrange(len(noisemat)):
                if i!=j:
                    idxil.append(i); idxjl.append(j)
                #|
            #|
        #|
        nondiagidx = (array(idxil),array(idxjl))
        m = 2*len(noisemat)*(len(noisemat)-1)
        S = lm[nondiagidx].sum()/m
        return exp(S)
    
